
package regression.data;

import java.util.Enumeration;
import java.util.NoSuchElementException;


/**
 * Arithmetic function.
 *
 * Parses and evaluates arithmetic expressions defined by the following grammar:
 *
 * 		Start		->	Expression
 *
 * 		Expression	->	x
 * 		Expression	->	( Expression Operator Expression )
 *
 * 		Operator	->	+
 * 		Operator	->	-
 * 		Operator	->	*
 * 		Operator	->	/
 *
 * @author	Copyright (C) 2004 Alessandro Bollini
 */
public class Function {

	private final Expression expression;
	private final double range=1e15; // allowed floating point range.


	public Function(final String expression) {

		if ( expression == null ) {
			throw new IllegalArgumentException("null expression");
		}

		this.expression=parse(expression);
	}


	public double evaluate(final double x) {
		return expression.evaluate(new Context(x));
	}


	private Expression parse(final String expression) {
		return parse(new Enumeration() {

			private int n=0;

			public boolean hasMoreElements() {
				return n < expression.length();
			}

			public Object nextElement() {
				if ( hasMoreElements() ) {
					return new Character(expression.charAt(n++)).toString();
				} else {
					throw new NoSuchElementException("no more elements");
				}
			}
		});
	}

	private Expression parse(final Enumeration scanner) {
		try {

			final Expression expression=parseExpression(scanner);

			if ( scanner.hasMoreElements() ) {
				throw new IllegalArgumentException(
						"unexpected tokens after end of expression");
			}

			return expression;

		} catch ( final NoSuchElementException exception ) {
			throw new IllegalArgumentException("unexpected end of expression");
		}
	}

	private Expression parseExpression(final Enumeration scanner) {

		final String token=(String)scanner.nextElement();

		if ( token.equals("x") ) {

			return new Variable();

		} else if ( token.equals("(") ) {

			final Expression x=parseExpression(scanner);
			final Operator operator=parseOperator(scanner);
			final Expression y=parseExpression(scanner);

			if ( !scanner.nextElement().equals(")") ) {
				throw new IllegalArgumentException("unexpected token "+token);
			}

			return new Operation(operator, x, y);

		} else {
			throw new IllegalArgumentException("unknow token "+token);
		}
	}

	private Operator parseOperator(final Enumeration scanner) {

		final String token=(String)scanner.nextElement();

		if ( token.equals("+") ) {
			return new SumOperator();

		} else if ( token.equals("-") ) {
			return new DifferenceOperator();

		} else if ( token.equals("*") ) {
			return new ProductOperator();

		} else if ( token.equals("/") ) {
			return new RatioOperator();

		} else {
			throw new IllegalArgumentException("unknow operator "+token);
		}
	}


	public String toString() {
		return expression.toString();
	}


	private interface Expression {
		public double evaluate(final Context context);
	}

	private class Context {

		private final double value;

		public Context(final double value) {
			this.value=value;
		}

		public double getValue() {
			return value;
		}
	}

	private class Variable implements Expression {

		public double evaluate(final Context context) {
			return context.getValue();
		}

		public String toString() {
			return "x";
		}
	}

	private class Operation implements Expression {

		private final Operator operator;
		private final Expression x, y;

		public Operation(
				final Operator operator,
				final Expression xExpression,
				final Expression yExpression) {

			this.operator=operator;
			this.x=xExpression;
			this.y=yExpression;
		}

		public double evaluate(final Context context) {
			return operator.process(
					x.evaluate(context),
					y.evaluate(context)
			);
		}

		public String toString() {
			return "("+x+operator+y+")";
		}
	}


	private interface Operator {
		public double process(final double x, final double y);
	}

	private class SumOperator implements Operator {

		public double process(final double x, final double y) {
			return clip(clip(x)+clip(y));
		}

		public String toString() {
			return "+";
		}
	}

	private class DifferenceOperator implements Operator {

		public double process(final double x, final double y) {
			return clip(clip(x)-clip(y));
		}

		public String toString() {
			return "-";
		}
	}

	private class ProductOperator implements Operator {

		public double process(final double x, final double y) {
			return clip(clip(x)*clip(y));
		}

		public String toString() {
			return "*";
		}
	}

	private class RatioOperator implements Operator {

		public double process(final double x, final double y) {
			return Math.abs(y) > 1.0/range ? clip(clip(x)/clip(y))
					: x > 0 ? range : x < 0 ? -range : 0.0;
		}

		public String toString() {
			return "/";
		}
	}

	private final double clip(final double x) {
		return x > range ? range : x < -range ? -range : x;
	}
}
