
package regression.data;

import java.io.IOException;
import java.util.*;


/**
 * Context-free grammar.
 *
 * @author  Copyright &copy; 2004 Alessandro Bollini. All rights reserved.
 */
public final class Grammar {

	static private final String start="Start";

	private final Map productions=new HashMap(); // <String, List<String[]>>


	public Grammar(final String file) throws IOException {

		final Tokenizer tokenizer=new Tokenizer(file);

		for (String[] tokens; (tokens=tokenizer.readTokens()) != null;) {

			if ( tokens.length < 3 || !tokens[1].equals("->") ) {
				throw new RuntimeException(
						"syntax error at line "+tokenizer.getLineNumber());
			}

			final String symbol=tokens[0];
			final String[] expansion=new String[tokens.length-2];

			for (int n=0; n < expansion.length; n++) {
				expansion[n]=tokens[n+2];
			}

			addProduction(symbol, expansion);
		}

		if ( !isSymbol(start) ) {
			throw new RuntimeException("symbol <Start> is not defined");
		}
	}

	private void addSymbol(final String symbol) {

		if ( symbol == null ) {
			throw new IllegalArgumentException("null symbol");
		}

		if ( !productions.containsKey(symbol) ) {
			productions.put(symbol, new ArrayList());
		}
	}

	private void addProduction(final String symbol, final String[] expansion) {

		if ( symbol == null ) {
			throw new IllegalArgumentException("null symbol");
		}

		if ( expansion == null ) {
			throw new IllegalArgumentException("null expansion");
		}

		addSymbol(symbol);

		for (int n=0; n < expansion.length; n++) {
			addSymbol(expansion[n]);
		}

		getExpansions(symbol).add(expansion);
	}

	private List getExpansions(final String symbol) {

		if ( symbol == null ) {
			throw new IllegalArgumentException("null symbol");
		}

		final List expansions=(List)productions.get(symbol);

		if ( expansions == null ) {
			throw new IllegalArgumentException("unknown symbol <"+symbol+">");
		}

		return expansions;
	}


	public boolean isSymbol(final String symbol) {

		if ( symbol == null ) {
			throw new IllegalArgumentException("null symbol");
		}

		return productions.containsKey(symbol);
	}

	public boolean isTerminal(final String symbol) {

		if ( symbol == null ) {
			throw new IllegalArgumentException("null symbol");
		}

		return isSymbol(symbol) && options(symbol) == 0;
	}


	public String start() {
		return start;
	}

	public int options(final String symbol) {
		return getExpansions(symbol).size();
	}

	public String[] expand(final String symbol, final int option) {
		return (String[])getExpansions(symbol).get(option);
	}
}
